@@ -21,20 +21,20 @@ import (
 	"path/filepath"
 	"testing"
 
-	"helm.sh/helm/v3/pkg/helmpath"
-	"helm.sh/helm/v3/pkg/helmpath/xdg"
+	"helm.sh/helm/v4/pkg/helmpath"
+	"helm.sh/helm/v4/pkg/helmpath/xdg"
 )
 
 // HelmHome sets up a Helm Home in a temp dir.
 func HelmHome(t *testing.T) {
 	t.Helper()
 	base := t.TempDir()
-	os.Setenv(xdg.CacheHomeEnvVar, base)
-	os.Setenv(xdg.ConfigHomeEnvVar, base)
-	os.Setenv(xdg.DataHomeEnvVar, base)
-	os.Setenv(helmpath.CacheHomeEnvVar, "")
-	os.Setenv(helmpath.ConfigHomeEnvVar, "")
-	os.Setenv(helmpath.DataHomeEnvVar, "")
+	t.Setenv(xdg.CacheHomeEnvVar, base)
+	t.Setenv(xdg.ConfigHomeEnvVar, base)
+	t.Setenv(xdg.DataHomeEnvVar, base)
+	t.Setenv(helmpath.CacheHomeEnvVar, "")
+	t.Setenv(helmpath.ConfigHomeEnvVar, "")
+	t.Setenv(helmpath.DataHomeEnvVar, "")
 }
 
 // TempFile ensures a temp file for unit testing purposes.
@@ -46,9 +46,10 @@ func HelmHome(t *testing.T) {
 //	tempdir := TempFile(t, "foo", []byte("bar"))
 //	filename := filepath.Join(tempdir, "foo")
 func TempFile(t *testing.T, name string, data []byte) string {
+	t.Helper()
 	path := t.TempDir()
 	filename := filepath.Join(path, name)
-	if err := os.WriteFile(filename, data, 0755); err != nil {
+	if err := os.WriteFile(filename, data, 0o755); err != nil {
 		t.Fatal(err)
 	}
 	return path
